/* Pinpad firmware for ogarev.

  Copyright (C) 2017, 2018 Ineiev <ineiev@gnu.org>

  This file is part of ogarev, a pinpad for Gnuk.

  Pinpad firmware is free software: you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Gnuk is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

/* Low fuse bits are 0xe2 (disable DIV8),
   high fuse bits should be 0xdf to make sure EEPROM is erased with Chip Erase,
   extended fuse bits don't matter for ATmega48 (default is 0xff),
             may be default (0xf9) for ATmega88/168;
   set lock bits to 0xfc to disable reading from chip memory.

   Internal 8MHz oscillator is used.  */
#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include <config.h>
#include <led-bits.h>
#include <display-signs.h>
#include <string.h>

static uint32_t
left_rotate (uint32_t x, unsigned char n)
{
  return ((x << n)) | (x >> (32 - n));
}
static void
quater_round (uint32_t* y,
              unsigned char i0, unsigned char i1,
              unsigned char i2, unsigned char i3)
{
  y[i1] ^= left_rotate(y[i0] + y[i3], 7);
  y[i2] ^= left_rotate(y[i1] + y[i0], 9);
  y[i3] ^= left_rotate(y[i2] + y[i1], 13);
  y[i0] ^= left_rotate(y[i3] + y[i2], 18);
}
static void
row_round (uint32_t* y)
{
  quater_round (y, 0, 1, 2, 3);
  quater_round (y, 5, 6, 7, 4);
  quater_round (y, 10, 11, 8, 9);
  quater_round (y, 15, 12, 13, 14);
}
static void
column_round (uint32_t* y)
{
  quater_round (y, 0, 4, 8, 12);
  quater_round (y, 5, 9, 13, 1);
  quater_round (y, 10, 14, 2, 6);
  quater_round (y, 15, 3, 7, 11);
}
static void
double_round(uint32_t *a)
{
  column_round(a);
  row_round(a);
}
static void
salsa20_hash (uint32_t* h, const uint32_t* x)
{
  unsigned char i;

  memmove (h, x, 64);
  for (i = 0; i < 10; i++)
    double_round (h);
  for (i = 0; i < 16; i++)
    h[i] += x[i];
}

struct salsa20_seq
{
  uint32_t out[16]; uint32_t data[16]; uint8_t i;
};

static struct salsa20_seq salsa_seq;

static const uint32_t sigma[4] =
{
  101 + (120ul << 8) + (112ul << 16) + ( 97ul << 24),
  110 + (100ul << 8) + ( 32ul << 16) + ( 51ul << 24),
   50 + ( 45ul << 8) + ( 98ul << 16) + (121ul << 24),
  116 + (101ul << 8) + ( 32ul << 16) + (107ul << 24)
};
/* Initialize sequence.
   seq -- structure to initialize;
   key -- 32-byte key;
   n   -- 8-byte nonce; */
static void
salsa20_init (struct salsa20_seq* seq, uint32_t* key, uint32_t* n)
{
  seq->i = 0;
  seq->data[0] = sigma[0];
  memmove (seq->data + 1, key, sizeof (*key) * 4);
  seq->data[5] = sigma[1];
  memmove (seq->data + 6, n, sizeof (*n) * 2);
  memset (seq->data + 8, 0, sizeof (*n) * 2);
  seq->data[10] = sigma[2];
  memmove (seq->data + 11, key + 4, sizeof (*key) * 4);
  seq->data[15] = sigma[3];
}
/* Generate next byte of salsa20 sequence.  */
static uint8_t
salsa20_gen (struct salsa20_seq* seq)
{
  uint8_t ret, i = seq->i;

  if (!i)
    {
      salsa20_hash (seq->out, seq->data);
      seq->data[8]++;
      if (!seq->data[8])
        seq->data[9]++;
    }
  ret = (seq->out[i >> 2] >> ((i & 0x3) << 3));
  seq->i = (i + 1) & 63;
  return ret;
}

static uint32_t salsa_key[8];
static uint32_t salsa_id;
static uint32_t salsa_iv[2];

#define EEPROM_KEY_ADDR ((void*)0)
#define EEPROM_ID_ADDR (((unsigned char*)(EEPROM_KEY_ADDR)) + sizeof (salsa_key))
#define EEPROM_TRY_ADDR (EEPROM_ID_ADDR + sizeof (salsa_id))
#define TRY_MASK (7)

enum d_pins
{
  TXD = 1,
  LE = 5,
  CLK = 6,
  SDI = 7
};
enum b_pins
{
  MISO = 4,
  COL0 = 5
};
enum c_pins
{
  COL1 = 0,
  COL2 = 1,
  ROW0 = 2,
  ROW1 = 3,
  ROW2 = 4,
  ROW3 = 5
};
/*  Segment layout:
     AA
    F  B
    F  B
     GG
    E  C
    E  C
     DD  H
*/
#define DIGITS(a) \
  LED_A##a | LED_B##a | LED_C##a | LED_D##a | LED_E##a | LED_F##a, \
  LED_B##a | LED_C##a, \
  LED_A##a | LED_B##a | LED_G##a | LED_D##a | LED_E##a, \
  LED_A##a | LED_B##a | LED_C##a | LED_D##a | LED_G##a, \
  LED_B##a | LED_C##a | LED_F##a | LED_G##a, \
  LED_A##a | LED_C##a | LED_D##a | LED_F##a | LED_G##a, \
  LED_A##a | LED_C##a | LED_D##a | LED_E##a | LED_F##a | LED_G##a, \
  LED_A##a | LED_B##a | LED_C##a | (LED_7_HAVE_F? LED_F##a: 0), \
  LED_A##a | LED_B##a | LED_C##a | LED_D##a | LED_E##a | LED_F##a | LED_G##a, \
  LED_A##a | LED_B##a | LED_C##a | LED_D##a | LED_F##a | LED_G##a, \
  LED_A##a | LED_B##a | LED_C##a | LED_E##a | LED_F##a | LED_G##a, \
  LED_C##a | LED_D##a | LED_E##a | LED_F##a | LED_G##a, \
  LED_A##a | LED_D##a | LED_E##a | LED_F##a, \
  LED_B##a | LED_C##a | LED_D##a | LED_E##a | LED_G##a, \
  LED_A##a | LED_D##a | LED_E##a | LED_F##a | LED_G##a, \
  LED_A##a | LED_E##a | LED_F##a | LED_G##a

#define DIGIT_MASK(a) \
  (LED_A##a | LED_B##a | LED_C##a | LED_D##a \
   | LED_E##a | LED_F##a | LED_G##a)

static const unsigned digits[2][16] = { { DIGITS(1) }, { DIGITS(2) } };

static void
set_leds (unsigned x)
{
  unsigned char i, d;

  for (i = 0; i < 16; i++)
    {
      d = PORTD;
      if (x & 1)
        d |= (1 << SDI);
      else
        d &= ~(1 << SDI);
      x >>= 1;
      PORTD = d;
      PORTD = d | (1 << CLK);
      PORTD = d;
    }
  PORTD ^= 1 << LE; PORTD ^= 1 << LE;
}
static unsigned keys;
static unsigned control;
static signed char state; /* 0 - waiting for command;
                             1 - entering password;
                            -1 - displaying error code.  */
static unsigned char
probe (void)
{
  static unsigned k;
  static unsigned char col;
  unsigned char r;

  if (!(TIFR0 & (1 << TOV0)))
    return 0;

  TIFR0 = 1 << TOV0;

  r = PINC & 0x3C;

  k <<= 4; k |= r;
  switch (col)
    {
      case 0: PORTC = 0; PORTB = 1 << COL0; break;
      case 1: PORTB = 0; PORTC = 1 << COL1; break;
      default: PORTC = 1 << COL2;
    }
  col++;
  if (col < 3)
    return 0;
  keys = k >> 2;
  col = 0;
  k = 0;
  return 1;
}
static signed char
get_key (void)
{
  unsigned char i;
  unsigned k = keys;
  static const char matrix[12] = {2, 5, 8, 0, 1, 4, 7, 10, 3, 6, 9, 11};

  for (i = 0; i < sizeof(matrix); i++, k >>= 1)
    if (k & 1)
      return matrix[i];
  return -1;
}
static void
init_salsa (void)
{
  eeprom_read_block (salsa_key, EEPROM_KEY_ADDR, sizeof (salsa_key));
  eeprom_read_block (&salsa_id, EEPROM_ID_ADDR, sizeof (salsa_id));
  salsa20_init (&salsa_seq, salsa_key, salsa_iv);
}
static char
get_byte (unsigned char* s)
{
  unsigned char c, r;

  while (!(UCSR0A & (1 << RXC0)));
  r = UDR0;
  if (r & 0xF0)
    return !0;
  while (!(UCSR0A & (1 << RXC0)));
  c = UDR0;
  if (c & 0xF0)
    return !0;
  *s = (r | (c << 4)) ^ salsa20_gen (&salsa_seq);
  return 0;
}
static void
reset_try_counter (void)
{
  unsigned char tc = eeprom_read_byte (EEPROM_TRY_ADDR);

  if (tc == 0xFF)
    return;
  eeprom_write_byte (EEPROM_TRY_ADDR, 0xFF);
}
static void
reset_key (void)
{
  unsigned char *sk = (unsigned char*) salsa_key;
  unsigned char i;

  for (i = 0; i < sizeof (salsa_key); i++)
    sk[i] = 0xFF;
  eeprom_write_block (sk, EEPROM_KEY_ADDR, sizeof (salsa_key));
  eeprom_write_block (sk, EEPROM_ID_ADDR, sizeof (salsa_id));
  reset_try_counter ();
}
static void
decrement_try_counter (void)
{
  unsigned char tc = eeprom_read_byte (EEPROM_TRY_ADDR);

  tc--;
  if (tc & TRY_MASK)
    eeprom_write_byte (EEPROM_TRY_ADDR, tc);
  else
    reset_key ();
  control = digits[0][tc & TRY_MASK];
}
static unsigned
check_retry_counter (void)
{
  unsigned char tc = eeprom_read_byte (EEPROM_TRY_ADDR) & TRY_MASK;

  if (tc == TRY_MASK)
    return LED_INIT;
  return digits[1][tc] | (LED_INIT & DIGIT_MASK(1));
}
/* Receive new key twice; if the copies coincide, write the new key to EEPROM
   and reset try counter.
   Returns 0 (when succeeds) or error codes sent by process_key_request
   (2 to 4).  */
static unsigned char
set_new_key (void)
{
  unsigned char buffer[sizeof (salsa_key) + sizeof (salsa_id)];
  unsigned char bis[sizeof (buffer)];
  unsigned char i;

  for (i = 0; i < sizeof (buffer); i++)
    if (get_byte (buffer + i))
      return 1;
  for (i = 0; i < sizeof (buffer); i++)
    if (get_byte (bis + i))
      return 2;
  if (memcmp (bis, buffer, sizeof (buffer)))
    return 3;
  eeprom_write_block (buffer, EEPROM_KEY_ADDR, sizeof (buffer));
  eeprom_read_block (bis, EEPROM_KEY_ADDR, sizeof (buffer));
  if (memcmp (bis, buffer, sizeof (buffer)))
    return 4;
  reset_try_counter ();
  return 0;
}
static void
push_char (unsigned char c)
{
  while (!(UCSR0A & (1 << UDRE0)));
  UDR0 = c ^ salsa20_gen (&salsa_seq);
}
static void
send_ack (unsigned char* buffer)
{
  unsigned char i;

  for (i = sizeof (salsa_iv); i < 16; i++)
    push_char (buffer[i]);
}
/* Check if buffer contains first bytes of current key.
   If yes, set a new key; if no, decrement try counter; when it's zero,
   reset EEPROM to initial (unprogrammed) state.
   Sends an ack: error code.
   Error codes:
     0 - success
     1 - failed to receive new key
     2 - failed to receive its copy
     3 - new key and its copy don't match
     4 - failed to write new key to EEPROM.  */
static void
process_key_request (unsigned char* buffer)
{
  unsigned char c;

  c = set_new_key ();
  control = digits[1][c & 0xF];
  set_leds (control);
  push_char (c);
  if (c)
    state = -1;
}
static unsigned char
check_uart (void)
{
  static unsigned char buffer[20];
  static signed char i = -1;
  unsigned char c, ret = !0;

  if (!(UCSR0A & (1 << RXC0)))
    return 0;
  c = UDR0;
  if (c & 0xF0)
    {
      i = 0;
      return 0;
    }
  if (i < 0)
    return 0;
  if (i & 1)
    buffer[i >> 1] |= c << 4;
  else
    buffer[i >> 1] = c;
  i++;
  if (i < sizeof (buffer) << 1)
    return 0;

  memcpy (salsa_iv, buffer, sizeof (salsa_iv));
  init_salsa ();
  for (i = sizeof (salsa_iv); i < sizeof (buffer); i++)
    buffer[i] ^= salsa20_gen (&salsa_seq);

  if (memcmp (buffer + 16, &salsa_id, sizeof (salsa_id)))
    {
      i = -1;
      decrement_try_counter ();
      return 0;
    }
  c = buffer[sizeof (salsa_iv)];
  state = 1;
  if (c == 0xFF)
    {
      i = -1;
      send_ack (buffer);
      process_key_request (buffer);
      return 0;
    }
  switch ((c >> 4) & 0xF)
    {
      case 0: control = LED_PIN; break;
      case 1: control = LED_RC; break;
      case 2: control = LED_ADMIN; break;
      case 3: control = LED_CHECK; break;
      default:
        control = state = 0;
        return 0;
    }
  switch (c & 0xF)
    {
      case 0: control = check_retry_counter (); ret = state = 0;
        break;
      case 1: control |= LED_CURRENT; break;
      case 2: control |= LED_NEW; break;
      case 3: control |= LED_CONFIRM; break;
      default:
        control = state = 0;
        return 0;
    }
  send_ack (buffer);
  reset_try_counter ();
  return ret;
}
static void
cycle (void)
{
  static unsigned prev_keys;
  static unsigned char n;
  unsigned x;
  signed char k;

  PORTB ^= (1 << MISO); /* Let the user know if this pin is alive.  */

  if (check_uart ()) /* A request to enter a PIN.  */
    n = 0;
  if (!probe ())
    return;

  if (TEST_KEYS_AND_LEDS)
    x = digits[0][get_key() & 0xF] | digits[1][(get_key() >> 4) & 0xF];
  else
    x = control;

  if (keys)
    x |= LED_H1;
  if (!n)
    x |= LED_H2;
  if (!x)
    x = LED_H2;
  set_leds (x);
  if (state <= 0)
    return;
  k = 0xFF;
  if (!prev_keys && keys)
    k = get_key ();
  prev_keys = keys;
  push_char (k);
  if (k & 0x40)
    return;
  if (k == 11)
    control = state = 0;
  else if (k == 10)
    {
      if (!n)
        control = state = 0;
      n = 0;
    }
  else
    n++;
}
static void
init_io (void)
{
  DDRD = (1 << LE) | (1 << CLK) | (1 << SDI) | (1 << TXD);
  DDRB = (1 << COL0) | (1 << MISO);
  DDRC = (1 << COL1) | (1 << COL2);
}
static void
init_timer (void)
{
  TCCR0B = (1 << CS02) | (1 << CS00); /* clk / 1024 period 32ms @ 8 MHz */
}

#define RATE (F_CPU / BAUD_RATE / 16 - 1)
#if RATE >= 1 << 16
# error UART baud rate is too low
#endif
#if F_CPU / (RATE + 1) / 16 - BAUD_RATE > BAUD_RATE / 15 \
   || BAUD_RATE - F_CPU / (RATE + 1) / 16 > BAUD_RATE / 15
# error UART baud error is too high
#endif
static void
init_uart (void)
{
  UBRR0H = (RATE >> 8) & 0xFF;
  UBRR0L = RATE & 0xFF;
  UCSR0C = (1 << UCSZ00) | (1 << UCSZ01);
  UCSR0B = (1 << TXEN0) | (1 << RXEN0);
}
static void
init_all (void)
{
  init_io ();
  init_timer ();
  init_uart ();
  init_salsa ();
}
int
main (void)
{
  init_all ();
  while (1)
    cycle ();
  return 0;
}
